# -*- coding: utf-8 -*-
#
# sathub/comum/util.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
import random
import sys

from satcfe.base import _FuncoesSAT
from satcfe.base import DLLSAT

from .config import conf

NUM_SESSAO_MIN = 100000
NUM_SESSAO_MAX = 999999

NUM_CAIXA_MIN = 0
NUM_CAIXA_MAX = 999

NUM_CAIXAS_MAX = (NUM_CAIXA_MAX - NUM_CAIXA_MIN) + 1

NUM_POR_FAIXA = int((NUM_SESSAO_MAX - NUM_SESSAO_MIN) / NUM_CAIXAS_MAX)


def faixa_numeracao(numero_caixa):
    """
    Retorna a faixa de numeração de sessão para o número do caixa.

    .. sourcecode:: python

        >>> a, b = faixa_numeracao(0) # caixa 0
        >>> NUM_SESSAO_MIN <= a <= NUM_SESSAO_MAX
        True
        >>> NUM_SESSAO_MIN <= b <= NUM_SESSAO_MAX
        True

        >>> x, y = faixa_numeracao(999) # caixa 999
        >>> NUM_SESSAO_MIN <= x <= NUM_SESSAO_MAX
        True
        >>> NUM_SESSAO_MIN <= y <= NUM_SESSAO_MAX
        True

    """
    _num_caixa = numero_caixa + 1
    n_min = ((_num_caixa * NUM_POR_FAIXA) + NUM_SESSAO_MIN) - NUM_POR_FAIXA
    n_max = (_num_caixa * NUM_POR_FAIXA) + NUM_SESSAO_MIN
    if _num_caixa > 1:
       n_min += 1
    return n_min, n_max


class NumeradorSessaoPorCaixa(object):
    """
    Um numerador de sessões persistente, gravando os números gerados em
    arquivos no formato JSON.

    Para evitar problemas de concorrência no acesso ao arquivo e para anular a
    possibilidade de números de sessão conflitantes entre os caixa, cada
    caixa terá o seu próprio arquivo de númeração de sessão e sua própria
    faixa de numeração, usando o seguinte esquema:

    A ER SAT deixa claro, no atributo B14 (``numeroCaixa``) que são possíveis
    caixas númerados entre 0 e 999, o que dá a possibilidade de 1000 caixas.

    Os números de sessão devem ser números de 6 dígitos, o que dá um limite
    entre 100000 e 999999. Assim, 999999-100000 dá 899999 que, se dividido
    por 1000 (caixas) dá uma faixa de cerca de 899:

    .. sourcecode:: python

        >>> 999999 - 100000
        899999
        >>> 899999 / 1000
        899

    Assim, cada caixa terá a sua própria faixa de numeração de sessão, o que
    torna impossível conflitos de número de sessão. Além disso, cada caixa
    terá um arquivo próprio, onde os números gerados serão persistidos,
    garantindo que os últimos 100 números gerados para aquele caixa não serão
    repetidos.

    +-------+-----------------------------+
    | Caixa | Faixa de Números de Sessão  |
    +=======+=============================+
    |   0   | 100000 a 100899             |
    +-------+-----------------------------+
    |   1   | 100900 a 101798             |
    +-------+-----------------------------+
    |                 ...                 |
    +-------+-----------------------------+
    |  999  | 998102 a 999000             |
    +-------+-----------------------------+

    """

    def __init__(self, tamanho=100, numero_caixa=1):
        super(NumeradorSessaoPorCaixa, self).__init__()
        self._memoria = []
        self._tamanho = tamanho
        self._numero_caixa = numero_caixa

        assert 0 <= self._numero_caixa <= 9, \
                'Numero do caixa fora da faixa (0..999): {}'.format(
                        self._numero_caixa)

        self._arquivo_json = os.path.join(conf.pasta_projeto,
                'sessoes-cx-{}.json'.format(self._numero_caixa))


    def __call__(self, *args, **kwargs):
        while True:
            numero = random.randint(*faixa_numeracao(self._numero_caixa))
            if numero not in self._memoria:
                self._memoria.append(numero)
                if len(self._memoria) > self._tamanho:
                    self._memoria.pop(0) # remove o mais antigo
                break
        self._escrever_memoria()
        return numero


    def _carregar_memoria(self):
        self._memoria[:] = []
        if os.path.exists(self._arquivo_json):
            with open(self._arquivo_json, 'r') as f:
                self._memoria = json.load(f)

        assert isinstance(self._memoria, list), \
                "Memoria de numeracao de sessao deve ser um objeto 'list'; "\
                "obtido {}".format(type(self._memoria))


    def _escrever_memoria(self):
        with open(self._arquivo_json, 'w') as f:
            json.dump(self._memoria, f)


def memoize(fn):
    # Implementação de memoize obtida de "Thumbtack Engineering"
    # https://www.thumbtack.com/engineering/a-primer-on-python-decorators/
    stored_results = {}

    def memoized(*args):
        try:
            # try to get the cached result
            return stored_results[args]
        except KeyError:
            # nothing was cached for those args. let's fix that.
            result = stored_results[args] = fn(*args)
            return result

    return memoized


@memoize
def instanciar_numerador_sessao(numero_caixa):
    return NumeradorSessaoPorCaixa(numero_caixa=numero_caixa)


@memoize
def instanciar_funcoes_sat(numero_caixa):
    funcoes_sat = _FuncoesSAT(
            dll=DLLSAT(
                    caminho=conf.caminho_dll,
                    convencao=conf.convencao_chamada),
            numerador_sessao=instanciar_numerador_sessao(numero_caixa))
    return funcoes_sat


def hexdump(data):
    def _cut(sequence, size):
        for i in xrange(0, len(sequence), size):
            yield sequence[i:i+size]
    _hex = lambda seq: ['{0:02x}'.format(b) for b in seq]
    _chr = lambda seq: [chr(b) if 32 <= b <= 126 else '.' for b in seq]
    raw_data = map(ord, data)
    hexpanel = [' '.join(line) for line in _cut(_hex(raw_data), 16)]
    chrpanel = [''.join(line) for line in _cut(_chr(raw_data), 16)]
    hexpanel[-1] = hexpanel[-1] + (chr(32) * (47 - len(hexpanel[-1])))
    chrpanel[-1] = chrpanel[-1] + (chr(32) * (16 - len(chrpanel[-1])))
    return '\n'.join('%s  %s' % (h, c) for h, c in zip(hexpanel, chrpanel))



# -*- coding: utf-8 -*-
#
# sathub/resources/trocarcodigodeativacao.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
    Este método permite a troca do código de ativação, que é uma senha,
    nomalmente definida pelo contribuinte proprietário do equipamento SAT.

    Se o código de ativação original tiver sido perdido, então o código de
    ativação de emergência deverá ser escrito no arquivo de configurações e
    deverá ser usada a opção ``2`` para que seja reconhecido pelo equipamento
    SAT como o código de ativação de emergência.

    Será muito mais prático usar o software de gerenciamento do fabricante para
    efetuar a troca do código de ativação, a menos que o estabelecimento siga
    uma política de troca frequente do código de ativação devido a questões de
    segurança. Neste caso, seria muito mais interessante implementar esta
    funcionalidade na aplicação de retaguarda, já que é uma operação puramente
    administrativa. Mas isso é apenas uma sugestão.
"""

import logging

from flask import abort
from flask.ext import restful

from ..comum.util import hexdump
from ..comum.util import instanciar_funcoes_sat
from ..custom import request_parser


logger = logging.getLogger('sathub.resource')


def opcao_codigo(value, name):
    # 1 - codigo de ativação normal
    # 2 - código de ativação de emergência
    opcao = int(value)
    if opcao not in [1, 2]: # FIXME: deveriam ser constantes em `satcomum`
        raise ValueError('Parametro {} invalido: {}'.format(name, value))
    return opcao


parser = request_parser()

parser.add_argument('opcao',
        type=opcao_codigo,
        required=True,
        help=u'Identifica a que se refere o código de ativação usado '
                u'para realizar a troca')

parser.add_argument('novo_codigo',
        type=str,
        required=True,
        help=u'Novo código de ativação')

parser.add_argument('novo_codigo_confirmacao',
        type=str,
        required=True,
        help=u'Confirmação do novo código de ativação')


class TrocarCodigoDeAtivacao(restful.Resource):

    def post(self):
        args = parser.parse_args()

        numero_caixa = args['numero_caixa']
        opcao = args['opcao']
        novo_codigo = args['novo_codigo']
        novo_codigo_confirmacao = args['novo_codigo_confirmacao']

        abort(501) # FIXME: consultar número de sessão não implementado
        # 501 - Server does not recognise method or lacks ability to fulfill
        # NOTE: https://github.com/base4sistemas/satcfe

        fsat = instanciar_funcoes_sat(numero_caixa)
        retorno = fsat.trocar_codigo_ativacao(
                opcao,
                novo_codigo,
                novo_codigo_confirmacao)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Retorno "TrocarCodigoDeAtivacao" ('
                    'numero_caixa={!r}, opcao={!r}, novo_codigo={!r}, '
                    'novo_codigo_confirmacao={!r})\n{}'.format(
                            numero_caixa,
                            opcao,
                            novo_codigo,
                            novo_codigo_confirmacao,
                            hexdump(retorno)))

        return dict(funcao='TrocarCodigoDeAtivacao', retorno=retorno)



# -*- coding: utf-8 -*-
#
# sathub/custom.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask.ext.restful import reqparse

from .comum.util import NUM_CAIXA_MIN
from .comum.util import NUM_CAIXA_MAX


def numero_caixa(value, name):
    numero = int(value)
    if NUM_CAIXA_MIN <= numero <= NUM_CAIXA_MAX:
        return numero
    raise ValueError('Parametro {} nao indica um numero de caixa '
            'valido: {}'.format(name, value))


parser = reqparse.RequestParser()

parser.add_argument('numero_caixa',
        type=numero_caixa,
        required=True,
        location=['headers', 'values'],
        help=u'O número do caixa de onde a solicitação partiu')


def request_parser():
    global parser
    return parser.copy()



# -*- coding: utf-8 -*-
#
# sathub/api.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask.ext import restful

from sathub import app

from sathub.resources.comunicarcertificadoicpbrasil import ComunicarCertificadoICPBRASIL
from sathub.resources.enviardadosvenda import EnviarDadosVenda
from sathub.resources.cancelarultimavenda import CancelarUltimaVenda
from sathub.resources.consultarsat import ConsultarSAT
from sathub.resources.testefimafim import TesteFimAFim
from sathub.resources.consultarstatusoperacional import ConsultarStatusOperacional
from sathub.resources.consultarnumerosessao import ConsultarNumeroSessao
from sathub.resources.configurarinterfacederede import ConfigurarInterfaceDeRede
from sathub.resources.associarassinatura import AssociarAssinatura
from sathub.resources.atualizarsoftwaresat import AtualizarSoftwareSAT
from sathub.resources.extrairlogs import ExtrairLogs
from sathub.resources.bloquearsat import BloquearSAT
from sathub.resources.desbloquearsat import DesbloquearSAT
from sathub.resources.trocarcodigodeativacao import TrocarCodigoDeAtivacao

api = restful.Api(app)

api.add_resource(ComunicarCertificadoICPBRASIL, '/hub/v1/comunicarcertificadoicpbrasil')
api.add_resource(EnviarDadosVenda, '/hub/v1/enviardadosvenda')
api.add_resource(CancelarUltimaVenda, '/hub/v1/cancelarultimavenda')
api.add_resource(ConsultarSAT, '/hub/v1/consultarsat')
api.add_resource(TesteFimAFim, '/hub/v1/testefimafim')
api.add_resource(ConsultarStatusOperacional, '/hub/v1/consultarstatusoperacional')
api.add_resource(ConsultarNumeroSessao, '/hub/v1/consultarnumerosessao')
api.add_resource(ConfigurarInterfaceDeRede, '/hub/v1/configurarinterfacederede')
api.add_resource(AssociarAssinatura, '/hub/v1/associarassinatura')
api.add_resource(AtualizarSoftwareSAT, '/hub/v1/atualizarsoftwaresat')
api.add_resource(ExtrairLogs, '/hub/v1/extrairlogs')
api.add_resource(BloquearSAT, '/hub/v1/bloquearsat')
api.add_resource(DesbloquearSAT, '/hub/v1/desbloquearsat')
api.add_resource(TrocarCodigoDeAtivacao, '/hub/v1/trocarcodigodeativacao')



# -*- coding: utf-8 -*-
#
# sathub/comum/config.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import logging
import os
import sys

from unidecode import unidecode

import satcomum.constantes


_CONF_DEBUG = 'debug'
_CONF_CAMINHO_DLL = 'caminho_dll'
_CONF_CONVENCAO_CHAMADA = 'convencao_chamada'
_CONF_CODIGO_ATIVACAO = 'codigo_ativacao'


_PADRAO_DEBUG = True
_PADRAO_CAMINHO_DLL = 'sat.dll'
_PADRAO_CONVENCAO_CHAMADA = satcomum.constantes.WINDOWS_STDCALL
_PADRAO_CODIGO_ATIVACAO = '123456789'


CONFIGURACAO_PADRAO = {
        _CONF_DEBUG: _PADRAO_DEBUG,
        _CONF_CAMINHO_DLL: _PADRAO_CAMINHO_DLL,
        _CONF_CONVENCAO_CHAMADA: _PADRAO_CONVENCAO_CHAMADA,
        _CONF_CODIGO_ATIVACAO: _PADRAO_CODIGO_ATIVACAO,
    }



class Configuracoes(object):

    def __init__(self):
        super(Configuracoes, self).__init__()

        self.pasta_projeto = os.path.abspath(os.path.join(
                os.path.abspath(os.path.dirname(__file__)), '..', '..'))

        self.arquivo = os.path.join(self.pasta_projeto, 'conf.json')
        self.debug = _PADRAO_DEBUG
        self.caminho_dll = _PADRAO_CAMINHO_DLL
        self.convencao_chamada = _PADRAO_CONVENCAO_CHAMADA
        self.codigo_ativacao = _PADRAO_CODIGO_ATIVACAO

        self._carregar()
        self._configurar_log()


    @property
    def nome_convencao_chamada(self):
        return [s for v,s in satcomum.constantes.CONVENCOES_CHAMADA \
                        if v == self.convencao_chamada][0]


    def _carregar(self):
        if os.path.exists(self.arquivo):
            with open(self.arquivo, 'r') as f:
                dados = json.load(f)
        else:
            dados = CONFIGURACAO_PADRAO
            # salva um arquivo de configurações para que possa ser editado
            with open(self.arquivo, 'w') as f:
                f.write(json.dumps(CONFIGURACAO_PADRAO, indent=4))

        self.debug = dados.get(_CONF_DEBUG, _PADRAO_DEBUG)
        self.codigo_ativacao = dados.get(
                _CONF_CODIGO_ATIVACAO, _PADRAO_CODIGO_ATIVACAO)

        self.caminho_dll = dados.get(_CONF_CAMINHO_DLL, _PADRAO_CAMINHO_DLL)
        self.caminho_dll = os.path.expanduser(
                self.caminho_dll.replace('/', os.path.sep))

        self.convencao_chamada = dados.get(
                _CONF_CONVENCAO_CHAMADA, _PADRAO_CONVENCAO_CHAMADA)

        _convencoes = [v for v,s in satcomum.constantes.CONVENCOES_CHAMADA]
        if self.convencao_chamada not in _convencoes:
            raise ValueError('Valor inesperado para convencao de '
                    'chamada: {}'.format(self.convencao_chamada))


    def _configurar_log(self):
        filename = os.path.join(self.pasta_projeto, 'sathub.log')

        handler = logging.FileHandler(filename)
        handler.setFormatter(
                logging.Formatter('%(asctime)s %(levelname)-8s [%(name)s] '
                        '(%(module)s:%(lineno)d) %(message)s'))

        level = logging.DEBUG if self.debug else logging.WARNING

        for name in ['flask', 'sathub',]:
            logger = logging.getLogger(name)
            logger.setLevel(level)
            logger.addHandler(handler)


    def descrever(self):
        """
        Descreve as configurações na saída padrão ou no terminal, se houver um.
        """
        import sathub

        _verbose(u'SATHub versão {}', sathub.__version__)
        _verbose(u'[-] Debug está {}', 'LIGADO' if self.debug else 'DESLIGADO')
        _verbose(u'[-] Caminho para DLL: {}', self.caminho_dll)
        _verbose(u'[-] Convenção de chamada para DLL: {}',
                self.nome_convencao_chamada)

        if not os.path.exists(self.caminho_dll):
            _verbose(u'** DLL NÃO ENCONTRADA **')

        _verbose(u'')


conf = Configuracoes()


def _verbose(message, *args):
    as_ascii = lambda s: unidecode(s) if isinstance(s, unicode) else s

    msg = message
    if args:
        msg = message.format(*args)

    if sys.stdout.isatty():
        sys.stdout.write('{}\n'.format(as_ascii(msg)))
        sys.stdout.flush()
    else:
        # requer "WSGIRestrictStdout Off"
        # https://code.google.com/p/modwsgi/wiki/DebuggingTechniques
        # https://code.google.com/p/modwsgi/wiki/ConfigurationDirectives#WSGIRestrictStdout
        if msg:
            print >> sys.stdout, 'SATHub: {}'.format(as_ascii(msg))



# -*- coding: utf-8 -*-
#
# setup.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

import sathub


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def read_requirements(filename, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    requirements = []
    with io.open(filename, encoding=encoding) as f:
        for line in f:
            if line.strip():
                requirements.append(line)
    return requirements


long_description = read('README.rst', 'CHANGES.rst')


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest # import here, cause outside the eggs aren't loaded
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
        name='sathub',
        version=sathub.__version__,
        description=u'Compartilhamento de Equipamentos SAT via RESTful API',
        long_description=long_description,
        packages=[
                'sathub',
                'sathub.comum',
                'sathub.resources',
            ],
        install_requires=read_requirements('requirements.txt'),
        extras_require={
                'testing': [
                        'pytest',
                        'pytest-cov',
                    ],
            },
        tests_require=read_requirements('test-requirements.txt'),
        cmdclass={
                'test': PyTest
            },
        # test_suite='sathub.tests',
        include_package_data=True,
        license='Apache Software License',
        platforms='any',
        url='http://github.com/base4sistemas/sathub/',
        author=u'Daniel Gonçalves',
        author_email='daniel@base4.com.br',
        classifiers = [
                'Development Status :: 1 - Planning',
                'Environment :: Other Environment',
                'Framework :: Flask',
                'Intended Audience :: Developers',
                'Intended Audience :: Information Technology',
                'License :: OSI Approved :: Apache Software License',
                'Natural Language :: Portuguese (Brazilian)',
                'Operating System :: OS Independent',
                'Programming Language :: Python',
                'Topic :: Office/Business :: Financial :: Point-Of-Sale',
            ]
    )




# -*- coding: utf-8 -*-
#
# sathub/resources/testefimafim.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from flask.ext import restful

from ..comum.util import hexdump
from ..comum.util import instanciar_funcoes_sat
from ..custom import request_parser


logger = logging.getLogger('sathub.resource')

parser = request_parser()

parser.add_argument('dados_venda',
        type=str,
        required=True,
        help=u'XML contendo os dados do CF-e de venda para o teste fim-a-fim')


class TesteFimAFim(restful.Resource):

    def post(self):
        args = parser.parse_args()

        numero_caixa = args['numero_caixa']
        dados_venda = args['dados_venda']

        fsat = instanciar_funcoes_sat(numero_caixa)
        retorno = fsat.teste_fim_a_fim(dados_venda)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Retorno "TesteFimAFim" '
                    '(numero_caixa=%s)\n%s', numero_caixa, hexdump(retorno))

        return dict(funcao='TesteFimAFim', retorno=retorno)
